<?php
/**
 * @copyright   Copyright (c) http://www.manadev.com
 * @license     http://www.manadev.com/license  Proprietary License
 */

namespace Manadev\ProductCollection\Resources;

use Magento\Framework\App\ObjectManager;
use Magento\Framework\DB\Select;
use Magento\InventorySalesApi\Api\Data\SalesChannelInterface;
use Magento\InventorySalesApi\Api\Data\SalesChannelInterfaceFactory;
use Magento\InventorySalesApi\Api\GetStockBySalesChannelInterface;
use Magento\Store\Api\StoreManagementInterface;
use Magento\Store\Model\StoreManagerInterface;
use Manadev\Core\Registries\ProductAttributes;
use Manadev\ProductCollection\Contracts\FacetResourceRegistry;
use Manadev\ProductCollection\Contracts\Filter;
use Manadev\ProductCollection\Contracts\FilterResourceRegistry;
use Manadev\ProductCollection\Contracts\QueryEngine;
use Manadev\ProductCollection\Query;
use Manadev\ProductCollection\Registries\FacetResources;
use Manadev\ProductCollection\Registries\FilterResources;
use Manadev\ProductCollection\Contracts\ProductCollection;
use Manadev\ProductCollection\Resources\HelperResource;
use Magento\Framework\ObjectManagerInterface;


class QueryEngineResource implements QueryEngine
{
    /**
     * @var FilterResources
     */
    protected $filterResources;
    /**
     * @var FacetResources
     */
    protected $facetResources;
    /**
     * @var \Manadev\ProductCollection\Resources\HelperResource
     */
    private $helperResource;
    /**
     * @var ObjectManagerInterface
     */
    private $objectManager;

    public function __construct(FilterResources $filterResources,
                                FacetResources $facetResources,
                                HelperResource $helperResource,
                                ObjectManagerInterface $objectManager)
    {
        $this->filterResources = $filterResources;
        $this->facetResources = $facetResources;
        $this->helperResource = $helperResource;
        $this->objectManager = $objectManager;
    }

    /**
     * @return FilterResourceRegistry
     */
    public function getFilterResourceRegistry()
    {
        return $this->filterResources;
    }

    /**
     * @return FacetResourceRegistry
     */
    public function getFacetResourceRegistry()
    {
        return $this->facetResources;
    }

    public function run(ProductCollection $productCollection)
    {
        $query = $productCollection->getQuery();

        $select = clone $productCollection->getSelect();

        $this->applyFiltersToSelect($productCollection->getSelect(), $query);

        foreach ($query->getFacets() as $facet) {
            $resource = $this->facetResources->get($facet->getType());

            if ($resource->isPreparationStepNeeded()) {
                $preparationSelect = $this->applyFiltersToSelect(clone $select, $query,
                    $resource->getPreparationFilterCallback($facet));
                $resource->prepare($preparationSelect, $facet);
            } else {
                $preparationSelect = null;
            }

            $facetSelect = $this->applyFiltersToSelect(clone $select, $query, $resource->getFilterCallback($facet));
            $facet->setData($resource->count($facetSelect, $facet));
        }
    }

    /**
     * @param Select $select
     * @param Filter $filter
     * @param null $callback
     * @return bool|false|string
     */
    public function applyFilterToSelectRecursively(Select $select, Filter $filter, $callback = null)
    {
        if ($callback && !call_user_func($callback, $filter)) {
            return false;
        }

        $resource = $this->filterResources->get($filter->getType());
        return $resource->apply($select, $filter, $callback);
    }

    protected function applyFiltersToSelect(Select $select, Query $query, $callback = null)
    {
        if ($condition = $this->applyFilterToSelectRecursively($select, $query->getFilters(), $callback)) {
            $select->where($condition);
        }

        /**
         *  FOS-545
         *
         * --  ________                          _____ _____
         * --  ___  __ \_____ __________________ __  /___  /_______
         * --  __  /_/ /_  _ \___  /___  /_  __ \_  __/_  __/_  __ \
         * --  _  ____/ /  __/__  /___  /_/ /_/ // /_  / /_  / /_/ /
         * --  /_/      \___/ _____/_____/\____/ \__/  \__/  \____/
         * --
         */

        if ($this->helperResource->isMultistockAvailable()) {
            $storeManager = $this->helperResource->getStore();
            $salesChannel = $this->objectManager->get(SalesChannelInterfaceFactory::class);
            $stockChannel = $this->objectManager->get(GetStockBySalesChannelInterface::class);

            $channel = $salesChannel->create();
            /** @var SalesChannelInterface $channel */
            $channel->setType(SalesChannelInterface::TYPE_WEBSITE);
            $channel->setCode($storeManager->getWebsite()->getCode());
            $stockInfo = $stockChannel->execute($channel);
            $select->join(
                ['inventory_stock_' . $stockInfo->getStockId()],
                'inventory_stock_' . $stockInfo->getStockId() . '.sku=e.sku AND inventory_stock_' . $stockInfo->getStockId() . '.is_salable > 0',
                []
            );

        } else {
            $select->join(
                ['cataloginventory_stock_item'],
                'cataloginventory_stock_item.product_id=e.entity_id AND cataloginventory_stock_item.is_in_stock > 0',
                []
            );
        }

        $sql = $select->__toString();

        return $select;
    }
}